/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011, 2012  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * Topological sort. 
 */

define schedule_test(data) = exists Start,Stop,Phase : {
    fixed_list(Start,|data[0]|) and
    fixed_list(Stop,|data[0]|) and
    make_schedule(|data[0]|,data[0],data[1],Start,Stop,Phase) and
    always format("Start=%t, Stop=%t\n",Start,Stop)
}.

define make_schedule(n,c,durations,Start,Stop,Phase) = {
    forall i<n : {
	waiting(i,c,Start,Stop);
	active(i,durations[i],Phase,Start,Stop);
	done(i,Start,Stop)
    } and
    Phase = 0 and
    Phase gets Phase+1 and
    halt (forall i<n : |Stop[i]|=1)
}.

define waiting(i,c,Start,Stop) = {
    Start[i] is [] and
    Stop[i] is [] and
    halt(forall j<|c[i]| : |Stop[c[i][j]]|=1)
}.

define active(i,t,Phase,Start,Stop) = {
    Start[i] <- [Phase] and
    fin{Stop[i] = [Phase]} and
    keep{
	Start[i]=[] and
	Stop[i]=[]
    } and
    len(t)
}.

define done(i,Start,Stop) = {
    stable (Start[i]) and
    stable (Stop[i])
}.

define keep_stable(X) = exists x : {x=X and keep(X=x)}.

define sort_process(i,a_list,Rdy_list) = {
    stable(Rdy_list[i]) and halt(forall j<|a_list[i]|:Rdy_list[a_list[i][j]]);
    skip and Rdy_list[i] := true; /* Don't use semicolon?? */
    stable(Rdy_list[i])
}.

define difference(L1,L2) = diff(|L1|-1,L1,L2).

define diff(i,L1,L2) = (if i<0 then []
		else diff(i-1,L1,L2) + (if L1[i]=L2[i] then [] else [i])).

define top_test1 = [[1],[2],[]].

define top_test2 = [[2],[7],[6],[5],[],[],[4,3],[5],[1,4]].

define top_test3 = [[[1],[],[0,3],[],[3]],[1,1,2,3,2]].

load "lists".
