/* -*- Mode: C -*-
 * This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011, 2012  Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
/* 
 * library of to ease interaction with verilog
 *
 */

define bits=[  1,         2,          4,          8,  
                        16,        32,         64,        128, 
                       256,       512,       1024,       2048,
                      4096,      8192,      16384,      32768,
                     65536,    131072,     262144,     524288,
                   1048576,   2097152,    4194304,    8388608,
                  16777216,  33554432,   67108864,  134217728,
                 268435456, 536870912, 1073741824 ].

define  bit (Bitno, Number)= {
    if ( (Number div (bits[Bitno])) mod 2 = 1) then 1 else 0}.

define nattobits(L,X) = {
    if L=1 then [bit(0,X)]
    else [bit(L-1,X)]+inttobits(L-1,X)
}.

define abs(X) = {
  if X<0 then -1*X else X
}.

define compl(X) = {
    if X=0 then 1 else 0
}.

define complement(L,X) = {
   if L=1 then compl(X[0])
   else [compl(X[L-1])]+complement(L-1,X)
}.
    

define neginttobits(X) = {
    complement(32,nattobits(32,abs(X)-1))
}.

define bitstonat(X) = {
    if |X| = 1 then X[0]
    else 
    {
        2**(|X|-1)*X[0] + bitstoint(suf(X,1))
    }
}.
/* declares an event variable X                   */
/* X[0] is used to memorise the old value         */
/* X[1] contains the current write value          */
/* X[2] is write flag for the event               */
/*        false: process i doesn't write          */
/*        true: process i does write              */
define event(X) = {
   list(X, 3) and stable(struct(X)) and
   /* memorise the current value of X */
   X[0] gets read_event(X)
}.

/* the current value of event                  X */
/* if there is no write then it is the old value */
/* otherwise it is the new value                 */
define read_event(X) = {
  if  X[2]=false  then X[0]
  else X[1]
}.

/* a posedge event */
define posedge_event(X) = {
    (X[2]=true and X[0]=0 and X[1]=1)
}.

/* a negedge event */
define negedge_event(X) = {
    (X[2]=true and X[0]=1 and X[1]=0)
}.

/* generate event X */
define write_event(X,Y) = {
  X[2]=true and X[1]=Y 
  and format("generate event %d by \n",Y)
}.

/* generate not event X */
define no_write_event(X) = {
 X[2]=false 
 and format("no event \n")
}.


/* run */ define test() = { empty and
  forall i<31 : format("bits %d = %t  \n",i,bits[i])
}.
