/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011, 2012  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: errors.c,v 2.15 2011/10/03 14:14:29 acau Exp $
 */



#include "system.h"


/*****************\
*                 *
* Error handling  *
*                 *
\*****************/

void pop_tempura_level(b)
int b;
{
    if (TempuraLevel > 0) {
        TempuraLevel -= 1;
        Break = b;
    }
    else
        if (b) {
            printf ("Type \"exit.\" to exit from Tempura.\n");
            fflush(stdout);
        }
    
}

void tempura_prefix(s)
char *s;
{
    printf ("\n***Tempura %s:  ", s);
}

/* print out an error message and break */
void error_break2(text,v1,v2)
char *text, *v1, *v2;
{
    tempura_prefix ("error");
    printf (text, v1, v2);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}

/* print out an error message and break */
void error_break2d(text,v1,v2)
char *text;
int v1, v2;
{
    tempura_prefix ("error");
    printf (text, v1, v2);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}

/* print out an error message and break */
void error_break2e(text,v1)
char *text;
int v1;
{
    tempura_prefix ("error");
    printf (text, v1);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}
/* print out a message and break */
void user_break(text)
char *text;
{
    tempura_prefix ("break");
    printf ("%s",text);
    printf (".\n");
    fflush(stdout);
    show_form ();
    breakin ();
}

/* print out a warning */
void warn1(text,v1)
char *text, *v1;
{
    tempura_prefix ("warning");
    printf (text, v1);
    printf (".\n");
    fflush(stdout);
    show_form ();
}

/* print out a warning with two data items*/
void warn2(text,v1,v2)
char *text, *v1, *v2;
{
    tempura_prefix ("warning");
    printf (text, v1, v2);
    printf (".\n");
    fflush(stdout);
    show_form ();
}



void show_state()
{
  int env_level;
  unsigned int i;
  newline;
  printf ("!Begin state: %d",StateNumber);
  newline;
  fflush(stdout);
  for (env_level = EnvTop; env_level >= EnvBase; env_level -= 1) {
      NODEPTR id = node_env_idlist(EnvStack[env_level].env);
      if (id != NULL)
          printf("Show_state: history=%d\n", node_env_history(EnvStack[env_level].env));
      
     while (id != NULL) {
         NODEPTR loc = curr_loc (mem_loc (id));
         output_str ("**Current");
         output_str (";");
         output_str (node_nm_name(id));
         output_str ("=");
         if (!mem_defined1 (loc)) {
                output_str ("?");
         }
         else {
             if (val_defined (get_mem (loc))) {
                     output_val(get_mem (loc));
                 }
             else {
                 output_str ("?");
             }
         }
         newline;
         loc = next_loc (mem_loc (id));
         output_str ("**Next");
         output_str (";");
         output_str (node_nm_name(id));
         output_str ("=");
         if (!mem_defined1 (loc)) {
                output_str ("?");
         }
         else {
             if (val_defined (get_mem (loc))) {
                     output_val(get_mem (loc));
                 }
             else {
                 output_str ("?");
             }
         }
         newline;
         for (i=2; i<=1+node_mem_history(loc); i=i+1) {
             loc = prev_loc1(mem_loc (id),i);
             printf ("%s%d","**Previous",node_mem_history(loc)+2-i);
             output_str (";");
             output_str (node_nm_name(id));
             output_str ("=");    
             if (!mem_defined1 (loc)) {
                output_str ("?");
             }
             else {
                 if (val_defined (get_mem (loc))) {
                     output_val(get_mem (loc));
                 }
                 else {
                     output_str ("?");
                 }
             }
             newline;
         }
         
         fflush(stdout);
         id = node_next(id);
     }
  }
  printf ("!End state: %d",StateNumber);
  newline;
  fflush(stdout);
}

         
         

void show_undef_vars()
{
    int env_level;
    char c;
    printf ("   Undefined: ");
    if (!empty_defined ()) {
        output_str ("{ empty }");
    }
    for (env_level = EnvTop; env_level >= EnvBase; env_level -= 1) {
        NODEPTR id = node_env_idlist(EnvStack[env_level].env);
	output_char ('{');
	c = ' ';
        while (id != NULL) {
            NODEPTR loc = curr_loc (mem_loc (id));
            if (!mem_defined1 (loc) || !val_defined (get_mem (loc))) {
                output_char (c);
                output_str (node_nm_name(id));
		c = ',';
	    }
	    id = node_next(id);
        }
        output_str (" }");
    }
    newline;
    fflush(stdout);
}

void state_not_defined()
{
    tempura_prefix ("error");
    printf ("state #%d (pass #%d) is not completely defined.\n",
             StateNumber, PassNumber);
    fflush(stdout);
    show_form ();
    show_undef_vars ();
    breakin ();
}

void show_form()
{
    NODEPTR e;
    if ((e = ReduceStackTop -> n) != NULL) {
        printf ("   Evaluating:  ");
        output_expn (e, PrintLevel);
        newline;
        fflush(stdout);
    }
}

void show_trace()
{
    int i=1;
    REDUCESTACKPTR rs;
    printf ("\nTrace of reductions:\n\n");
    fflush(stdout);
    for (rs = ReduceStack; rs < ReduceStackTop; rs += 1) {
        printf ("%3d. ",i++);
        output_expn (rs -> n, PrintLevel);
        newline;
        fflush(stdout);
    }
    newline;
    fflush(stdout);
}

void show_scope()
{
    int env_level;
    printf ("Environment:\n\n");
    fflush(stdout);
    for (env_level = EnvTop; env_level >= EnvBase; env_level -= 1) {
        printf ("%3d. ", node_env_level(EnvStack[env_level].env));
        output_env (EnvStack[env_level].env);
        newline;
        fflush(stdout);
    }
    newline;
    fflush(stdout);
}
