/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011, 2012  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: globals.h,v 2.15 2011/10/03 14:14:29 acau Exp $
 */

#include "fns.h"

extern unsigned Ready;
extern unsigned Retry;
extern unsigned NextFlag;
extern unsigned int PrevFlag;
extern unsigned RedefFlag;
extern NODEPTR EmptyLoc;
extern unsigned Clock;
extern unsigned AlarmGone;
extern unsigned IsInitial;


/* counter for top-level commands */
extern unsigned Counter;

/* Memory */
extern NODE GlobalMemory[];
extern unsigned LocalMemUsed;
extern NODEPTR TempLoc;
extern unsigned CurrBit;
extern unsigned NextBit;


extern unsigned OneLevelMemFlag;
extern unsigned TimeStampCheckFlag;

/* environment */
extern ENVSTACK EnvStack[];
extern int EnvTop;
extern int EnvBase;

/* reduce-stack */
extern REDUCESTACK ReduceStack[];
extern REDUCESTACKPTR ReduceStackTop;
extern REDUCESTACKPTR ReduceStackMax;
extern NODEPTR LastCommand;

/* tracing and debugging */
extern jmp_buf *BreakPoint;
extern int TempuraLevel;
extern int BreakDepth;
extern unsigned Break;
extern unsigned InterruptSet;
extern int PassLimitCounter;
extern unsigned PassProgressFlag;
extern unsigned StateProgressFlag;
extern unsigned ForceOutput;
extern unsigned RunFlag;
extern int SyntaxErrorCount;

/* statistics */
extern unsigned StateNumber;
extern unsigned PassNumber;
extern unsigned TotalPassCount;
extern unsigned TotalReduceCount;
extern unsigned TotalReduceTimes;
extern unsigned StateReduceCount;
extern unsigned PassReduceCount;
extern unsigned TotalSuccessfulReduceCount;
extern unsigned TotalSuccessfulReduceTimes;
extern unsigned StateSuccessfulReduceCount;
extern unsigned PassSuccessfulReduceCount;
extern int ReduceDepth;
extern int MaxReduceDepth;

/* Static values (very commonly used) */
extern NODEPTR True, False;
extern NODEPTR Zero, One;
extern NODEPTR Nil, Empty, More, Skip, Ranlen, Infinite, Random, RRandom, SRandom, SRRandom;
extern NODEPTR EndOfFile;

/* Heap */
extern HEAP Heap;
extern unsigned HeapSize;
extern unsigned FreeNodes;
extern NODEPTR FreeNodeList;
extern NODEPTR TempNodeList;

/* Symbol Table */
extern NODEPTR SymbolTableTop;
extern NODEPTR SymbolTableMax;
extern NODEPTR HashTable[];

/* Wait list for system calls */
extern NODEPTR SysWaitList;

/* System variables (user definable). */
extern int PassLimit;
extern int Verbosity;
extern int PrintLevel;
extern int DebugLevel;
extern unsigned StatePrintFlag;
extern unsigned RedundancyFlag;
extern unsigned CallByNameFlag;
extern unsigned AssignAheadFlag;
extern unsigned BreakIsAbortFlag;
extern unsigned HopChop;

/* To fix the path problem:
   extern char *Path;
   extern char *HelpDir;
*/
extern NODEPTR Path;
extern NODEPTR HelpDir;
extern NODEPTR InFile;
extern NODEPTR OldInFile;
extern NODEPTR OutFile;
extern FILE *Fpcur;
extern char *InFilename;

extern int TimeStep;
/* AC 29Apr1999 */
extern long MaxRandLen;

extern long RandSeed;
extern int TRandSeed;
extern long RandomVal;
extern unsigned RandomValTime;

extern long RRandSeed;
extern int TRRandSeed;
extern long RRandomVal;
extern unsigned RRandomValTime;

extern long SRandSeed;
extern int TSRandSeed;
extern double SRandomVal;
extern unsigned SRandomValTime;

extern long SRRandSeed;
extern int TSRRandSeed;
extern double SRRandomVal;
extern unsigned SRRandomValTime;

/* floats */
extern int Precision;
/* Input/Output */
extern NODEPTR AuxVal;
extern int CharsOutput;
extern int InputLineCount;
extern FILE *InputFile;
extern char InputText[];
extern char *InputFilename;

/* Parser */
extern unsigned Token;
extern unsigned Recover;
extern char *Keyword[];

/* Reduction functions */

extern PFI Reduces[MAXREDUCETOKEN + 1][3];
extern char *Reduc[MAXREDUCETOKEN + 1][3];
/*extern PFI ReduceStmt[];
extern PFI ReduceVal[];
extern PFI ReduceRef[];
*/
