/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011, 2012  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: init.c,v 2.18 2011/10/03 14:14:30 acau Exp $
 */


#include "system.h"





/*******************\
*                   *
* GLOBAL VARIABLES  *
*                   *
\*******************/

/* Static values (very commonly used) */
struct node TrueNode = {T,BOOL,NULL,{.bval=T}};
NODEPTR True = &TrueNode;
struct node FalseNode = {T,BOOL,NULL,{.bval=F}};
NODEPTR False = &FalseNode;
struct node ZeroNode = {T,INT,NULL,{.ival=0}};
NODEPTR Zero = &ZeroNode;
struct node OneNode = {T,INT,NULL,{.ival=1}};
NODEPTR One = &OneNode;
struct node NilNode = {T,NIL,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Nil = &NilNode;
struct node EmptyNode = {T,EMPTY,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Empty = &EmptyNode;
struct node MoreNode = {T,MORE,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR More = &MoreNode;
struct node SkipNode = {T,SKIP,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Skip = &SkipNode;
struct node RanlenNode = {T,RANLEN,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Ranlen = &RanlenNode;
struct node InfiniteNode = {T,INFINITE,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Infinite = &InfiniteNode;
struct node RandomNode = {T,RANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR Random = &RandomNode;
struct node RRandomNode = {T,RRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR RRandom = &RRandomNode;
struct node SRandomNode = {T,SRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR SRandom = &SRandomNode;
struct node SRRandomNode = {T,SRRANDOM,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR SRRandom = &SRRandomNode;
struct node EndOfFileNode = {T,END,NULL,{.lnkval={NULL,NULL,NULL}}};
NODEPTR EndOfFile = &EndOfFileNode;
/*MEMORY GlobalEmpty[MEMCELLSIZE] = {{T,0,T,F,0,nodep (&TrueNode)},*/
/* 			           {T,1,T,F,0,nodep (&TrueNode)}};*/
NODE GlobalEmpty[MEMCELLSIZE];
NODEPTR EmptyLoc = GlobalEmpty;

unsigned Ready = F;                  /* True after successful reduction. */
unsigned Retry = F;                  /* Re-reduce in current pass? */
unsigned NextFlag = F;               /* Reduce next value of expression? */
unsigned int PrevFlag = 0; 
unsigned RedefFlag = F;              /* Redefinition allowed? */
unsigned Clock = 0;
unsigned AlarmGone = F;
unsigned IsInitial = T;
/* counter for top-level commands */
unsigned Counter = 1;

/* Memory */
NODE GlobalMemory[offset2 (GLOBALMEMSIZE,1)];
unsigned LocalMemUsed = 0;
/*MEMORY TempMem[MEMCELLSIZE] = {{T,0,T,F,0,NULL},*/
/* 			       {T,1,T,F,0,NULL}};*/
NODE TempMem[MEMCELLSIZE];
NODEPTR TempLoc = TempMem;

unsigned CurrBit = 0;
unsigned NextBit = 1;


unsigned OneLevelMemFlag = F;        /* if on, use one-level memory */
unsigned TimeStampCheckFlag = T;     /* if on, check time stamp */

/* environment */
NODE GlobalEnv = {T,ENV};
 
                         
ENVSTACK EnvStack[MAXENVSTACKSIZE] = {{GlobalMemory, &GlobalEnv}};
int EnvTop = 0;
int EnvBase = 0;

/* reduce-stack */
REDUCESTACK ReduceStack[MAXREDUCESTACKSIZE];
REDUCESTACKPTR ReduceStackTop = ReduceStack+(-1);
REDUCESTACKPTR ReduceStackMax = ReduceStack+(MAXREDUCESTACKSIZE-1);
NODEPTR LastCommand = &TrueNode;

/* tracing and debugging */
jmp_buf *BreakPoint;
int TempuraLevel = 0;
int BreakDepth = -1;
unsigned Break = F;
unsigned InterruptSet = F;
int PassLimitCounter = 0;
unsigned PassProgressFlag = F;
unsigned StateProgressFlag = T;
unsigned ForceOutput = F;
unsigned RunFlag = F;
int SyntaxErrorCount = 0;

/* statistics */
unsigned StateNumber = 0;
unsigned PassNumber = 0;
unsigned TotalPassCount = 0;
unsigned TotalReduceCount = 0;
unsigned TotalReduceTimes = 0;
unsigned StateReduceCount = 0;

unsigned PassReduceCount = 0;
unsigned TotalSuccessfulReduceCount = 0;
unsigned TotalSuccessfulReduceTimes = 0;
unsigned StateSuccessfulReduceCount = 0;
unsigned PassSuccessfulReduceCount = 0;
int ReduceDepth = 0;
int MaxReduceDepth = 0;

/* Heap */
HEAP Heap = NULL;
unsigned FreeNodes = 0;
unsigned HeapSize = 0;
NODEPTR FreeNodeList = NULL;
NODEPTR TempNodeList = NULL;

/* Symbol Table */
NODEPTR SymbolTableTop = 0;
NODEPTR SymbolTableMax = 0;
NODEPTR HashTable[HASHTABLESIZE];/* The hash table. */

/* Wait list for system calls */
NODEPTR SysWaitList = NULL;

/* System variables (user definable). */
int PassLimit = -1;
int Verbosity = 1;
int PrintLevel = 2;
int DebugLevel = 1;
int Precision = 5;
unsigned StatePrintFlag = T;
unsigned RedundancyFlag = F;
unsigned CallByNameFlag = F;
unsigned AssignAheadFlag = T;
unsigned BreakIsAbortFlag = T;
unsigned HopChop = T;

NODEPTR Path;       /* BCM 14Dec96 */
NODEPTR HelpDir;    /* BCM 14Dec96 */
NODEPTR InFile;
NODEPTR OldInFile;
NODEPTR OutFile;
FILE *Fpcur = NULL;
char *InFilename = "Standard Input";


/* Previous two lines were as follows.  Had problems with garbage collection.
                BCM 14Dec96
char *Path;
char *HelpDir = "Help";
*/
int TimeStep = 0;
/* values related to random number generator AC 29Apr1999*/
long MaxRandLen = 10;   /* Maximum random interval length. BCM 22Jul97 */

long RandSeed = 0;      /* Seed for random numbers. BCM 16Jul97 */
int TRandSeed = 0;
long RandomVal = -1;     /* random value >=0 for current state.
 			   Initially "undefined" as -1. BCM 15Aug97 */
unsigned RandomValTime = 0; /* Time of random value. BCM 15Aug97 */

long RRandSeed = 0;      /* Seed for Random numbers. AC 20/09/2000 */
int TRRandSeed = 0; 
long RRandomVal = -1;     /* Random value >=0.
 			   Initially "undefined" as -1. AC 20/09/2000 */
unsigned RRandomValTime = 0; /* Time of Random value. AC 20/09/2000 */

long SRandSeed = 0;      /* Seed for frandom numbers. AC 20/09/2000 */
int TSRandSeed = 0; 
double SRandomVal = -1;     /* frandom value in [0.0,1.0).
 			   Initially "undefined" as -1. AC 20/09/2000 */
unsigned SRandomValTime = 0; /* Time of frandom value. AC 20/09/2000 */

long SRRandSeed = 0;      /* Seed for fRandom numbers. AC 20/09/2000 */
int TSRRandSeed = 0;
double SRRandomVal = -1;     /* fRandom value in [0.0,1.0).
 			   Initially "undefined" as -1. AC 20/09/2000 */
unsigned SRRandomValTime = 0; /* Time of fRandom value. AC 20/09/2000 */
/* Input/Output */
NODEPTR AuxVal;
int CharsOutput = 0;
int InputLineCount = 1;
FILE *InputFile; /*={stdin};*/
char InputText[MAXTOKENSIZE];
char *InputFilename = "Standard Input";

/* AC 29Apr1999: added
   111 random
   112 random_seed
   AC 20/09/2000: added
   113 Random
   114 Random_seed
   115 max_rand_len
   AC 04/09/2001: added
   116 implies
   AC 05/09/2001: added
   117 more
   118 chopstar
   AC 06/09/2001: added
   119 _if
*/

/* Parser */
unsigned Token = END;
unsigned Recover = F;
char *Keyword[MAXTOKEN + 1] = {
/*  0 */                ".",
/*  1 */                "nil",
/*  2 */                "bool",
/*  3 */                "integer",
/*  4 */                "lambda",
/*  5 */                "string",
/*  6 */                "float",
/*  7 */                "list",
/*  8 */                "state",
/*  9 */                "static",
/* 10  */                "empty",
/* 11 */                "skip",
/* 12 */                "and",
/* 13 */                "_and",
/* 14 */                "or",
/* 15 */                "not",
/* 16 */                "if",
/* 17 */                "_call",
/* 18 */                "_call",
/* 19 */                "exists",
/* 20 */                "_exists",
/* 21 */                "process",
/* 22 */                "_process",
/* 23 */                "prefix",
/* 24 */                "_prefix",
/* 25 */                "forall",
/* 26 */                "the",
/* 27 */                "always",
/* 28 */                "sometimes",
/* 29 */                "halt",
/* 30 */                "init",
/* 31 */                "fin",
/* 32 */                "keep",
/* 33 */                "next",
/* 34 */                "wnext",
/* 35 */                "len",
/* 36 */                "_len",
/* 37 */                "<",
/* 38 */                "<=",
/* 39 */                "=",
/* 40 */                ">=",
/* 41 */                ">",
/* 42 */                "~=",
/* 43 */                "gets",
/* 44 */                "stable",
/* 45 */                "where",
/* 46 */                "is",
/* 47 */                "<-",
/* 48 */                ":=",
/* 49 */                ";",
/* 50 */                "_;",
/* 51 */                "while",
/* 52 */                "repeat",
/* 53 */                "for",
/* 54 */                "_for",
/* 55 */                "_for",
/* 56 */                "_for_times",
/* 57 */                "_for",
/* 58 */                "_for_in",
/* 59 */                "proj",
/* 60 */                "_proj",
/* 61 */                "input",
/* 62 */                "output",
/* 63 */                "+",
/* 64 */                "-",
/* 65 */                "*",
/* 66 */                "div",
/* 67 */                "mod",
/* 68 */                "**",
/* 69 */                "+",
/* 70 */                "-",
/* 71 */                "_make_list",
/* 72 */                "type",
/* 73 */                "struct",
/* 74 */                "_size",
/* 75 */                "_sub",
/* 76 */                "_sub_list",
/* 77 */                "list",
/* 78 */                "List",
/* 79 */                "ref",
/* 80 */                "deref",
/* 81 */                "loc",
/* 82 */                "run",
/* 83 */                "exit",
/* 84 */                "load",
/* 85 */                "define",
/* 86 */                "set",
/* 87 */                "showtrace",
/* 88 */                "showmem",
/* 89 */                "abort",
/* 90 */                "break",
/* 91 */                "return",
/* 92 */                "continue",
/* 93 */                "leap",
/* 94 */                "step",
/* 95 */                "peek",
/* 96 */                "format",
/* 97 */                "ascii",
/* 98 */                "help",
/* 99 */                "system",
/*100 */                "_system",
/*101 */		"pass_limit",
/*102 */		"verbosity",
/*103 */		"print_level",
/*104 */		"debug_level",
/*105 */		"print_states",
/*106 */		"no_redundancy",
/*107 */		"call_by_name",
/*108 */		"assign_ahead",
/*109 */                "break_is_abort",
/*110 */		"path",
/*111 */		"help_dir",
/*112 */		"time_step",
/*113 */                "random",
/*114 */                "random_seed",
/*115 */                "Random",
/*116 */                "Random_seed",
/*117 */                "max_randlen",
/*118 */                "implies",
/*119 */                "more",
/*120 */                "chopstar",
/*121 */                "_if",
/*122 */                "_undefined",
/*123 */                "showstate",
/*124 */                "infile",
/*125 */                "outfile",
/*126 */                "randlen",
/*127 */                "infinite",
/*128 */                "hopchop",
/*129 */                "precision",
/*130 */                "ceil",
/*131 */                "floor",
/*132 */                "sqrt",
/*133 */                "itof",
/*134 */                "exp",
/*135 */                "log",
/*136 */                "log10",
/*137 */                "sin",
/*138 */                "cos",
/*139 */                "tan",
/*140 */                "asin",
/*141 */                "acos",
/*142 */                "atan",
/*143 */                "atan2",
/*144 */                "sinh",
/*145 */                "cosh",
/*146 */                "tanh",
/*147 */                "fabs",
/*148 */                "frandom",
/*149 */                "frandom_seed",
/*150 */                "fRandom",
/*151 */                "fRandom_seed",
/*152 */                "_or",
/*153 */                "defaults",
/*154 */                "mem",
/*155 */                "prev",
/*156 */                "nprev",
/*157 */                "winput",
/*158 */                ",",
/*159 */                "true",
/*160 */                "false",
/*161 */                "_name",
/*162 */                "then",
/*163 */                "else",
/*164 */                "in",
/*165 */                "times",
/*166 */                "_env",
/*167 */                "_bnd",
/*168 */                "to",
/*169 */                ":",
/*170 */                "(",
/*171 */                ")",
/*172 */                "{",
/*173 */                "}",
/*174 */                "[",
/*175 */                "]",
/*176 */                "|",
/*177 */                "until",
/*178 */                "#",
/*179 */                "_memory"
};

/* AC 29Apr1999 added
   111 assert_expr
   112 assert_expr
   113 assert_expr
   AC 20/09/2000 added
   114 assert_expr
   115 assert_expr
   AC 04/09/2001 added
   116 implies
   AC 05/09/2001 added
   117 more
   118 chopstar
   AC 06/09/2001 added
   119 _if
*/
PFI Reduces[MAXREDUCETOKEN + 1][3] = {
/*  0 END */
    {reduce_EOF,          reduce_EOF,         reduce_EOF},
/*  1 NIL */
    {assert_expr,         reduce_const,       get_ref},
/*  2 BOOL */
    {assert_true,         reduce_const,       get_ref},
/*  3 INT */
    {assert_expr,         reduce_const,       get_ref},
/*  4 LAMBDA */
    {assert_expr,         reduce_const,       get_ref},
/*  5 STRING */
    {assert_expr,         reduce_const,       get_ref},
/*  6 FLOAT */
    {assert_expr,         reduce_const,       get_ref},    
/*  7 LIST */
    {assert_expr,         reduce_const,       get_ref},
/*  8 STATE */
    {assert_true,         reduce_id_val,      reduce_id_ref},
/*  9 STATIC */
    {assert_true,         reduce_id_val,      reduce_id_ref},
/* 10 EMPTY */
    {reduce_EMPTY_stmt,   reduce_EMPTY_val,   reduce_EMPTY_ref},
/* 11 SKIP */
    {reduce_SKIP,         assert_stmt,        assert_stmt},
/* 12 AND */
    {reduce_AND,          reduce_bin_op,      get_ref},
/* 13 iAND */
    {reduce_iAND,         assert_stmt,        assert_stmt},
/* 14 OR */
    {reduce_OR,           reduce_bin_op,      get_ref},
/* 15 NOT */
    {assert_true,         reduce_u_op,        get_ref},
/* 16 IF */
    {reduce_IF_stmt,      reduce_IF_val,      get_ref},
/* 17 CALL */
    {reduce_CALL,         reduce_CALL,        reduce_CALL},
/* 18 iCALL */
    {reduce_iCALL_stmt,   reduce_iCALL_val,   reduce_iCALL_ref},
/* 19 EXISTS */
    {reduce_EXISTS,       reduce_EXISTS,      reduce_EXISTS},
/* 20 iEXISTS */
    {reduce_iEXISTS_stmt, reduce_iEXISTS_val, get_ref},
/* 21 PROCESS */
    {reduce_PROCESS,      assert_stmt,        assert_stmt},
/* 22 iPROCESS */
    {reduce_iPROCESS,     assert_stmt,        assert_stmt},
/* 23 PREFIX */
    {reduce_PREFIX,       assert_stmt,        assert_stmt},
/* 24 iPREFIX */
    {reduce_iPREFIX,      assert_stmt,        assert_stmt},
/* 25 FORALL */
    {reduce_FORALL,       reduce_FORALL,      reduce_FORALL},
/* 26 THE */
    {assert_expr,         reduce_THE,         reduce_THE},
/* 27 ALWAYS */
    {reduce_ALWAYS,       assert_stmt,        assert_stmt},
/* 28 SOMETIME */
    {reduce_SOMETIME,     assert_stmt,        assert_stmt},
/* 29 HALT */
    {reduce_HALT,         assert_stmt,        assert_stmt},
/* 30 INIT */
    {reduce_INIT,         assert_stmt,        assert_stmt},
/* 31 FIN */
    {reduce_FIN,          assert_stmt,        assert_stmt},
/* 32 KEEP */
    {reduce_KEEP,         assert_stmt,        assert_stmt},
/* 33 NEXT */
    {reduce_NEXT_stmt,    reduce_NEXT_val,    reduce_NEXT_ref},
/* 34 WNEXT */
    {reduce_WNEXT,        assert_stmt,        assert_stmt},
/* 35 LEN */
    {reduce_LEN,          assert_stmt,        assert_stmt},
/* 36 iLEN */
    {reduce_iLEN,         assert_stmt,        assert_stmt},
/* 37 LT */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 38 LE */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 39 EQ */
    {reduce_EQ_stmt,      reduce_EQ_val,      get_ref},
/* 40 GE */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 41 GT */
    {assert_stmt_to_true, reduce_bin_op,      get_ref},
/* 42 NEQ */
    {assert_stmt_to_true, reduce_EQ_val,      get_ref},
/* 43 GETS */
    {reduce_GETS,         assert_stmt,        assert_stmt},
/* 44 STB */
    {reduce_STB,          assert_stmt,        assert_stmt},
/* 45 WHERE */
    {reduce_WHERE,        assert_stmt,        assert_stmt},
/* 46 IS */
    {reduce_IS,           assert_stmt,        assert_stmt},
/* 47 ASSN */
    {reduce_ASSN,         assert_stmt,        assert_stmt},
/* 48 UASSN */
    {reduce_UASSN,        assert_stmt,        assert_stmt},
/* 49 SEMI */
    {reduce_SEMI,         assert_stmt,        assert_stmt},
/* 50 iSEMI */
    {reduce_iSEMI,        assert_stmt,        assert_stmt},
/* 51 WHILE */
    {reduce_WHILE,        assert_stmt,        assert_stmt},
/* 52 REPEAT */
    {reduce_REPEAT,       assert_stmt,        assert_stmt},
/* 53 FOR */
    {reduce_FOR,          assert_stmt,        assert_stmt},
/* 54 iFOR */
    {assert_expr,         reduce_iFOR,        assert_stmt},
/* 55 FORTIMES */
    {reduce_FORTIMES,     assert_stmt,        assert_stmt},
/* 56 iFORTIMES */
    {assert_expr,         reduce_iFORTIMES,   assert_stmt},
/* 57 FORIN */
    {reduce_FORIN,        assert_stmt,        assert_stmt},
/* 58 iFORIN */
    {assert_expr,         reduce_iFORIN,      assert_stmt},
/* 59 PROJ */
    {reduce_PROJ,         assert_stmt,        assert_stmt},
/* 60 iPROJ */
    {reduce_iPROJ,        assert_stmt,        assert_stmt},
/* 61 INPUT */
    {reduce_INPUT,        assert_stmt,        assert_stmt},
/* 62 OUTPUT */
    {reduce_OUTPUT,       reduce_OUTPUT,      assert_stmt},
/* 63 PLUS */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 64 MINUS */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 65 MULT */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 66 DIV */
    {assert_expr,         reduce_bin_op,      get_ref},    
/* 67 MOD */
    {assert_expr,         reduce_bin_op,      get_ref},
/* 68 POWER */
    {assert_expr,         reduce_bin_op,      get_ref}, 
/* 69 UPLUS */
    {assert_expr,         reduce_u_op,        get_ref},
/* 70 UMINUS */
    {assert_expr,         reduce_u_op,        get_ref},
/* 71 MAKELIST */
    {assert_expr,         reduce_MAKELIST,    get_ref},
/* 72 TYPE */
    {assert_expr,         reduce_TYPE,        get_ref},
/* 73 STRUCT */
    {assert_expr,         reduce_STRUCT_val,  reduce_STRUCT_ref},
/* 74 SIZE */
    {assert_expr,         get_val,            reduce_SIZE},
/* 75 SUB */
    {assert_true,         get_val,            reduce_SUB},
/* 76 SUBLIST */
    {assert_expr,         get_val,            reduce_SUBLIST},
/* 77 FIXLIST */
    {reduce_list,         assert_stmt,        assert_stmt},
/* 78 VARLIST */
    {reduce_list,         assert_stmt,        assert_stmt},
/* 79 REFER */
    {assert_expr,         reduce_REFER,       get_ref},
/* 80 DEREF */
    {assert_expr,         get_val,            reduce_DEREF},
/* 81 LOC */
    {assert_expr,         reduce_const,       get_ref},
/* 82 RUN */
    {reduce_RUN,          assert_stmt,        assert_stmt},
/* 83 EXIT */
    {reduce_EXIT,         assert_stmt,        assert_stmt},
/* 84 LOAD */
    {reduce_LOAD,         assert_stmt,        assert_stmt},
/* 85 DEFINE */
    {reduce_DEFINE,       assert_stmt,        assert_stmt},
/* 86 SET */
    {reduce_SET,          assert_stmt,        assert_stmt},
/* 87 SHOWTRACE */
    {reduce_SHOWTRACE,    assert_stmt,        assert_stmt},
/* 88 SHOWMEM */
    {reduce_SHOWMEM,      assert_stmt,        assert_stmt},
/* 89 ABORT */
    {reduce_ABORT,        assert_stmt,        assert_stmt},
/* 90 BREAK */
    {reduce_BREAK,        assert_stmt,        assert_stmt},
/* 91 RETURN */
    {reduce_RETURN,       assert_stmt,        assert_stmt},
/* 92 CONTINUE */
    {reduce_CONTINUE,     assert_stmt,        assert_stmt},
/* 93 LEAP */
    {reduce_LEAP,         assert_stmt,        assert_stmt},
/* 94 STEP */
    {reduce_STEP,         assert_stmt,        assert_stmt},
/* 95 PEEK */
    {reduce_PEEK,         assert_stmt,        assert_stmt},
/* 96 FORMAT */
    {reduce_FORMAT,       assert_stmt,        assert_stmt},
/* 97 ASCII */
    {assert_expr,         reduce_ASCII,       get_ref},
/* 98 HELP */
    {reduce_HELP,         assert_stmt,        assert_stmt},
/* 99 SYSTEM */
    {reduce_SYSTEM,       assert_stmt,        assert_stmt},
/*100 iSYSTEM */
    {reduce_iSYSTEM,      assert_stmt,        assert_stmt},
/*101 PASSLIMIT */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*102 VERBOSITY */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*103 PRINTLEVEL */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*104 DEBUGLEVEL */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*105 STATEPRINT */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*106 REDUNDANCY */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*107 CALLBYNAME */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*108 EAD */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*109 BREAKISABORT */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*110 PATH */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*111 HELPDIR */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*112 TIMESTEP */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*113 RANDOM */
    {assert_expr,         reduce_RANDOM,      get_ref},
/*114 RANDSEED */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*115 RRANDOM */
    {assert_expr,         reduce_RRANDOM,     get_ref},
/*116 RRANDSEED */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*117 MAXRANDLEN */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*118 IMPLIES */
    {reduce_IMPLIES,      reduce_IMPLIES,     reduce_IMPLIES},
/*119 MORE */
    {reduce_MORE_stmt,    reduce_MORE_val,    get_ref},
/*120 CHOPSTAR */
    {reduce_CHOPSTAR,     assert_stmt,        assert_stmt},
/*121 iIF */
    {assert_expr,         reduce_iIF_val,     get_ref},
/*122 iUNDEFINED */
    {assert_expr,         reduce_iUNDEFINED,   get_ref},
/*123 SHOWSTATE */
    {reduce_SHOWSTATE,     assert_stmt,       assert_stmt},
/*124 INFILE */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*125 OUTFILE */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*126 RANLEN */
    {reduce_RANLEN,       assert_stmt,        assert_stmt},
/*127 INFINITE */
    {reduce_INFINITE,     assert_stmt,        assert_stmt},
/*128 HOPCHOP */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*129 PRECISION */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*130 CEIL */
    {assert_expr,         reduce_u_op,        get_ref},
/*131 FLOOR */
    {assert_expr,         reduce_u_op,        get_ref},
/*132 SQRT */
    {assert_expr,         reduce_u_op,        get_ref},
/*133 ITOF */
    {assert_expr,         reduce_u_op,        get_ref},
/*134 EXPO */
    {assert_expr,         reduce_u_op,        get_ref},
/*135 LOG */
    {assert_expr,         reduce_u_op,        get_ref},
/*136 LOGTEN */
    {assert_expr,         reduce_u_op,        get_ref},
/*137 SIN */
    {assert_expr,         reduce_u_op,        get_ref},
/*138 COS */
    {assert_expr,         reduce_u_op,        get_ref},
/*139 TAN */
    {assert_expr,         reduce_u_op,        get_ref},
/*140 ASIN */
    {assert_expr,         reduce_u_op,        get_ref},
/*141 ACOS */
    {assert_expr,         reduce_u_op,        get_ref},
/*142 ATAN */
    {assert_expr,         reduce_u_op,        get_ref},
/*143 ATANTWO */
    {assert_expr,         reduce_bin_op,      get_ref},
/*144 SINH */
    {assert_expr,         reduce_u_op,        get_ref},
/*145 COSH */
    {assert_expr,         reduce_u_op,        get_ref},
/*146 TANH */
    {assert_expr,         reduce_u_op,        get_ref},
/*147 FABS */
    {assert_expr,         reduce_u_op,        get_ref},
/*148 SRANDOM */
    {assert_expr,         reduce_SRANDOM,     get_ref},
/*149 SRANDSEED */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*150 SRRANDOM */
    {assert_expr,         reduce_SRRANDOM,    get_ref},
/*151 SRRANDSEED */
    {assert_expr,         reduce_SYSVAR,      get_ref},
/*152 iOR */
    {reduce_iOR,          assert_stmt,        assert_stmt},
/*153 DEFAULTS */
    {reduce_DEFAULTS,     assert_stmt,        assert_stmt},
/*154 MEM */
    {reduce_MEM,          assert_stmt,        assert_stmt},
/*155 PREV */
    {reduce_PREV_stmt,    reduce_PREV_val,    reduce_PREV_ref},
/*156 NPREV */
    {reduce_NPREV_stmt,   reduce_NPREV_val,   reduce_NPREV_ref},
/*157 WINPUT */
    {reduce_WINPUT,        assert_stmt,        assert_stmt},    
/*158 COMMA */
    {noop,                noop,               noop}
};


char *Reduc[MAXREDUCETOKEN + 1][3] = {
/*  0 END */
    {"reduce_EOF",          "reduce_EOF",         "reduce_EOF"},
/*  1 NIL */
    {"assert_expr",         "reduce_const",       "get_ref"},
/*  2 BOOL */
    {"assert_true",         "reduce_const",       "get_ref"},
/*  3 INT */
    {"assert_expr",         "reduce_const",       "get_ref"},
/*  4 LAMBDA */
    {"assert_expr",         "reduce_const",       "get_ref"},
/*  5 STRING */
    {"assert_expr",         "reduce_const",       "get_ref"},
/*  6 FLOAT */
    {"assert_expr",         "reduce_const",       "get_ref"},    
/*  7 LIST */
    {"assert_expr",         "reduce_const",       "get_ref"},
/*  8 STATE */
    {"assert_true",         "reduce_id_val",      "reduce_id_ref"},
/*  9 STATIC */
    {"assert_true",         "reduce_id_val",      "reduce_id_ref"},
/* 10 EMPTY */
    {"reduce_EMPTY_stmt",   "reduce_EMPTY_val",   "reduce_EMPTY_ref"},
/* 11 SKIP */
    {"reduce_SKIP",         "assert_stmt",        "assert_stmt"},
/* 12 AND */
    {"reduce_AND",          "reduce_bin_op",      "get_ref"},
/* 13 iAND */
    {"reduce_iAND",         "assert_stmt",        "assert_stmt"},
/* 14 OR */
    {"reduce_OR",           "reduce_bin_op",      "get_ref"},
/* 15 NOT */
    {"assert_true",         "reduce_u_op",        "get_ref"},
/* 16 IF */
    {"reduce_IF_stmt",      "reduce_IF_val",      "get_ref"},
/* 17 CALL */
    {"reduce_CALL",         "reduce_CALL",        "reduce_CALL"},
/* 18 iCALL */
    {"reduce_iCALL_stmt",   "reduce_iCALL_val",   "reduce_iCALL_ref"},
/* 19 EXISTS */
    {"reduce_EXISTS",       "reduce_EXISTS",      "reduce_EXISTS"},
/* 20 iEXISTS */
    {"reduce_iEXISTS_stmt", "reduce_iEXISTS_val", "get_ref"},
/* 21 PROCESS */
    {"reduce_PROCESS",      "assert_stmt",        "assert_stmt"},
/* 22 iPROCESS */
    {"reduce_iPROCESS",     "assert_stmt",        "assert_stmt"},
/* 23 PREFIX */
    {"reduce_PREFIX",       "assert_stmt",        "assert_stmt"},
/* 24 iPREFIX */
    {"reduce_iPREFIX",      "assert_stmt",        "assert_stmt"},
/* 25 FORALL */
    {"reduce_FORALL",       "reduce_FORALL",      "reduce_FORALL"},
/* 26 THE */
    {"assert_expr",         "reduce_THE",         "reduce_THE"},
/* 27 ALWAYS */
    {"reduce_ALWAYS",       "assert_stmt",        "assert_stmt"},
/* 28 SOMETIME */
    {"reduce_SOMETIME",     "assert_stmt",        "assert_stmt"},
/* 29 HALT */
    {"reduce_HALT",         "assert_stmt",        "assert_stmt"},
/* 30 INIT */
    {"reduce_INIT",         "assert_stmt",        "assert_stmt"},
/* 31 FIN */
    {"reduce_FIN",          "assert_stmt",        "assert_stmt"},
/* 32 KEEP */
    {"reduce_KEEP",         "assert_stmt",        "assert_stmt"},
/* 33 NEXT */
    {"reduce_NEXT_stmt",    "reduce_NEXT_val",    "reduce_NEXT_ref"},
/* 34 WNEXT */
    {"reduce_WNEXT",        "assert_stmt",        "assert_stmt"},
/* 35 LEN */
    {"reduce_LEN",          "assert_stmt",        "assert_stmt"},
/* 36 iLEN */
    {"reduce_iLEN",         "assert_stmt",        "assert_stmt"},
/* 37 LT */
    {"assert_stmt_to_true", "reduce_bin_op",      "get_ref"},
/* 38 LE */
    {"assert_stmt_to_true", "reduce_bin_op",      "get_ref"},
/* 39 EQ */
    {"reduce_EQ_stmt",      "reduce_EQ_val",      "get_ref"},
/* 40 GE */
    {"assert_stmt_to_true", "reduce_bin_op",      "get_ref"},
/* 41 GT */
    {"assert_stmt_to_true", "reduce_bin_op",      "get_ref"},
/* 42 NEQ */
    {"assert_stmt_to_true", "reduce_EQ_val",      "get_ref"},
/* 43 GETS */
    {"reduce_GETS",         "assert_stmt",        "assert_stmt"},
/* 44 STB */
    {"reduce_STB",          "assert_stmt",        "assert_stmt"},
/* 45 WHERE */
    {"reduce_WHERE",        "assert_stmt",        "assert_stmt"},
/* 46 IS */
    {"reduce_IS",           "assert_stmt",        "assert_stmt"},
/* 47 ASSN */
    {"reduce_ASSN",         "assert_stmt",        "assert_stmt"},
/* 48 UASSN */
    {"reduce_UASSN",        "assert_stmt",        "assert_stmt"},
/* 49 SEMI */
    {"reduce_SEMI",         "assert_stmt",        "assert_stmt"},
/* 50 iSEMI */
    {"reduce_iSEMI",        "assert_stmt",        "assert_stmt"},
/* 51 WHILE */
    {"reduce_WHILE",        "assert_stmt",        "assert_stmt"},
/* 52 REPEAT */
    {"reduce_REPEAT",       "assert_stmt",        "assert_stmt"},
/* 53 FOR */
    {"reduce_FOR",          "assert_stmt",        "assert_stmt"},
/* 54 iFOR */
    {"assert_expr",         "reduce_iFOR",        "assert_stmt"},
/* 55 FORTIMES */
    {"reduce_FORTIMES",     "assert_stmt",        "assert_stmt"},
/* 56 iFORTIMES */
    {"assert_expr",         "reduce_iFORTIMES",   "assert_stmt"},
/* 57 FORIN */
    {"reduce_FORIN",        "assert_stmt",        "assert_stmt"},
/* 58 iFORIN */
    {"assert_expr",         "reduce_iFORIN",      "assert_stmt"},
/* 59 PROJ */
    {"reduce_PROJ",         "assert_stmt",        "assert_stmt"},
/* 60 iPROJ */
    {"reduce_iPROJ",        "assert_stmt",        "assert_stmt"},
/* 61 INPUT */
    {"reduce_INPUT",        "assert_stmt",        "assert_stmt"},
/* 62 OUTPUT */
    {"reduce_OUTPUT",       "reduce_OUTPUT",      "assert_stmt"},
/* 63 PLUS */
    {"assert_expr",         "reduce_bin_op",      "get_ref"},
/* 64 MINUS */
    {"assert_expr",         "reduce_bin_op",      "get_ref"},
/* 65 MULT */
    {"assert_expr",         "reduce_bin_op",      "get_ref"},
/* 66 DIV */
    {"assert_expr",         "reduce_bin_op",      "get_ref"},    
/* 67 MOD */
    {"assert_expr",         "reduce_bin_op",      "get_ref"},
/* 68 POWER */
    {"assert_expr",         "reduce_bin_op",      "get_ref"}, 
/* 69 UPLUS */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/* 70 UMINUS */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/* 71 MAKELIST */
    {"assert_expr",         "reduce_MAKELIST",    "get_ref"},
/* 72 TYPE */
    {"assert_expr",         "reduce_TYPE",        "get_ref"},
/* 73 STRUCT */
    {"assert_expr",         "reduce_STRUCT_val",  "reduce_STRUCT_ref"},
/* 74 SIZE */
    {"assert_expr",         "get_val",            "reduce_SIZE"},
/* 75 SUB */
    {"assert_true",         "get_val",            "reduce_SUB"},
/* 76 SUBLIST */
    {"assert_expr",         "get_val",            "reduce_SUBLIST"},
/* 77 FIXLIST */
    {"reduce_list",         "assert_stmt",        "assert_stmt"},
/* 78 VARLIST */
    {"reduce_list",         "assert_stmt",        "assert_stmt"},
/* 79 REFER */
    {"assert_expr",         "reduce_REFER",       "get_ref"},
/* 80 DEREF */
    {"assert_expr",         "get_val",            "reduce_DEREF"},
/* 81 LOC */
    {"assert_expr",         "reduce_const",       "get_ref"},
/* 82 RUN */
    {"reduce_RUN",          "assert_stmt",        "assert_stmt"},
/* 83 EXIT */
    {"reduce_EXIT",         "assert_stmt",        "assert_stmt"},
/* 84 LOAD */
    {"reduce_LOAD",         "assert_stmt",        "assert_stmt"},
/* 85 DEFINE */
    {"reduce_DEFINE",       "assert_stmt",        "assert_stmt"},
/* 86 SET */
    {"reduce_SET",          "assert_stmt",        "assert_stmt"},
/* 87 SHOWTRACE */
    {"reduce_SHOWTRACE",    "assert_stmt",        "assert_stmt"},
/* 88 SHOWMEM */
    {"reduce_SHOWMEM",      "assert_stmt",        "assert_stmt"},
/* 89 ABORT */
    {"reduce_ABORT",        "assert_stmt",        "assert_stmt"},
/* 90 BREAK */
    {"reduce_BREAK",        "assert_stmt",        "assert_stmt"},
/* 91 RETURN */
    {"reduce_RETURN",       "assert_stmt",        "assert_stmt"},
/* 92 CONTINUE */
    {"reduce_CONTINUE",     "assert_stmt",        "assert_stmt"},
/* 93 LEAP */
    {"reduce_LEAP",         "assert_stmt",        "assert_stmt"},
/* 94 STEP */
    {"reduce_STEP",         "assert_stmt",        "assert_stmt"},
/* 95 PEEK */
    {"reduce_PEEK",         "assert_stmt",        "assert_stmt"},
/* 96 FORMAT */
    {"reduce_FORMAT",       "assert_stmt",        "assert_stmt"},
/* 97 ASCII */
    {"assert_expr",         "reduce_ASCII",       "get_ref"},
/* 98 HELP */
    {"reduce_HELP",         "assert_stmt",        "assert_stmt"},
/* 99 SYSTEM */
    {"reduce_SYSTEM",       "assert_stmt",        "assert_stmt"},
/*100 iSYSTEM */
    {"reduce_iSYSTEM",      "assert_stmt",        "assert_stmt"},
/*101 PASSLIMIT */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*102 VERBOSITY */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*103 PRINTLEVEL */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*104 DEBUGLEVEL */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*105 STATEPRINT */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*106 REDUNDANCY */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*107 CALLBYNAME */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*108 ASSIGNAHEAD */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*109 BREAKISABORT */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*110 PATH */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*111 HELPDIR */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*112 TIMESTEP */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*113 RANDOM */
    {"assert_expr",         "reduce_RANDOM",      "get_ref"},
/*114 RANDSEED */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*115 RRANDOM */
    {"assert_expr",         "reduce_RRANDOM",     "get_ref"},
/*116 RRANDSEED */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*117 MAXRANDLEN */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*118 IMPLIES */
    {"reduce_IMPLIES",      "reduce_IMPLIES",     "reduce_IMPLIES"},
/*119 MORE */
    {"reduce_MORE_stmt",    "reduce_MORE_val",    "get_ref"},
/*120 CHOPSTAR */
    {"reduce_CHOPSTAR",     "assert_stmt",        "assert_stmt"},
/*121 iIF */
    {"assert_expr",         "reduce_iIF_val",     "get_ref"},
/*122 iUNDEFINED */
    {"assert_expr",         "reduce_iUNDEFINED",   "get_ref"},
/*123 SHOWSTATE */
    {"reduce_SHOWSTATE",     "assert_stmt",       "assert_stmt"},
/*124 INFILE */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*125 OUTFILE */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*126 RANLEN */
    {"reduce_RANLEN",       "assert_stmt",        "assert_stmt"},
/*127 INFINITE */
    {"reduce_INFINITE",     "assert_stmt",        "assert_stmt"},
/*128 HOPCHOP */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*129 PRECISION */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*130 CEIL */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*131 FLOOR */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*132 SQRT */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*133 ITOF */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*134 EXPO */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*135 LOG */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*136 LOGTEN */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*137 SIN */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*138 COS */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*139 TAN */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*140 ASIN */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*141 ACOS */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*142 ATAN */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*143 ATANTWO */
    {"assert_expr",         "reduce_bin_op",      "get_ref"},
/*144 SINH */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*145 COSH */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*146 TANH */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*147 FABS */
    {"assert_expr",         "reduce_u_op",        "get_ref"},
/*148 SRANDOM */
    {"assert_expr",         "reduce_SRANDOM",     "get_ref"},
/*149 SRANDSEED */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*150 SRRANDOM */
    {"assert_expr",         "reduce_SRRANDOM",     "get_ref"},
/*151 SRRANDSEED */
    {"assert_expr",         "reduce_SYSVAR",      "get_ref"},
/*152 iOR */
    {"reduce_iOR",          "assert_stmt",        "assert_stmt"},
/*153 DEFAULTS */
    {"reduce_DEFAULTS",     "assert_stmt",        "assert_stmt"},
/*154 MEM */
    {"reduce_MEM",          "assert_stmt",        "assert_stmt"},
/*155 PREV */
    {"reduce_PREV_stmt",    "reduce_PREV_val",    "reduce_PREV_ref"},
/*156 NPREV */
    {"reduce_NPREV_stmt",   "reduce_NPREV_val",   "reduce_NPREV_ref"},
/*157 WINPUT */
    {"reduce_WINPUT",        "assert_stmt",        "assert_stmt"},    
/*158 COMMA */
    {"noop",                "noop",               "noop"}
};





/***********************************************\
*                                               *
*  Enter reserved words into the symbol table.  *
*                                               *
\***********************************************/

void install_keywords()
{
    int code;
    for (code = 1; code <= MAXTOKEN; code += 1) {
        if (isalpha((int)*Keyword[code]))
            install_keyword(code,Keyword[code]);
    }

    /* Alternative forms for operators. */

    install_keyword(DO,strsave("do"));
    install_keyword(ALWAYS,strsave("box"));
    install_keyword(SOMETIME,strsave("sometime"));
    install_keyword(LEAP,strsave("jump"));
    install_keyword(NEXT,strsave("shift"));
}


/* initialize tempura */
void tempura_init ()
{
    catch_signals ();
    install_keywords ();
    init_sysvars ();
    new_loc1 (TempLoc, STATIC, 1);
    put_mem1 (new_loc1 (EmptyLoc, STATIC, 1), True);
}
