/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011, 2012  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: lexer.c,v 2.14 2011/10/03 14:14:30 acau Exp $
 */


#include "system.h"



#define BS '\\'
#define ERROR 256
/*********************\
*                     *
*  Lexical analyser   *
*                     *
\*********************/

int read_token()
{
    register int     c;
    register char   *p = InputText;
    register char   *pmax = InputText + MAXTOKENSIZE - 1;
    char     *strsave ();

    AuxVal = EndOfFile;
    while (isspace (*p = c = getc (InputFile)) ||
    	   (c == '/' && isnextchar ('*'))) {
        if (c == '/') {         /* Discard comment. */
            do {
                while ((c = getc (InputFile)) != '*') {
                   if (c == EOF) {
                       syntax_error (IDN);
                       return (END);
                   }
                   if (c == '\n')
                        InputLineCount++;
                }
            } while (!isnextchar ('/'));
        }
        if (c == '\n')
            InputLineCount++;
    }
    InputText[1] = '\0';

    if (isalpha (c)) {
        do {
            *(p < pmax ? p++ : pmax) = c;
        } while ((c = getc (InputFile)) != EOF && (isalnum (c) || (c == '_')));
        ungetc (c, InputFile);
        *(p < pmax ? p : pmax) = '\0';
        if (p > pmax)
            warn1 ("identifier truncated: \"%s\"", InputText);
        if ((AuxVal = lookup (InputText)) == NULL) {
            AuxVal = install (NAME, strsave (InputText));
        }
        return (node_type(AuxVal));
    }

    if (isdigit (c)) {
        do {
            *(p < pmax ? p++ : pmax) = c;
        } while ((c = getc (InputFile)) != EOF && isdigit (c));
        ungetc (c, InputFile);
        *(p < pmax ? p : pmax) = '\0';
        AuxVal = ((c = atoi (InputText)) == 0) ?
            Zero :
            (c == 1) ?
            One :
            int_node (c);
        return (INT);
    }

    if (c == '"') {
        while ((c = getc (InputFile)) != '"') {
            if (c == EOF) {
                *p = '\0';
                syntax_error (MQS);
                return (END);
            }
            *(p < pmax ? p++ : pmax) = backslash (c);
        }
        *(p < pmax ? p : pmax) = '\0';
        if (p >= pmax)
            syntax_error (STL);
        AuxVal = string_node (strsave (InputText));
        return (STRING);
    }

    if (c == '\'') {
        while ((c = getc (InputFile)) != '\'') {
            if (c == EOF) {
                *p = '\0';
                syntax_error (MQS);
                return (END);
            }
            *(p < pmax ? p++ : pmax) = c;
        }
        *(p < pmax ? p : pmax) = '\0';
        if (p >= pmax)
            syntax_error (STL);
        AuxVal = string_node (strsave (InputText));
        return (STRING);
    }

    if (c == '$') {
        while ((c = getc (InputFile)) != '$') {
            if (c == EOF) {
                *p = '\0';
                syntax_error (MDF);
                return (END);
            }
            if (!isdigit(c) && c != '.' && c != 'e' && c != '+' && c != '-') {
                *p = '\0';
                syntax_error (EIF);
            }
            *(p < pmax ? p++ : pmax) = c;
        }
        *(p < pmax ? p : pmax) = '\0';
        if (p >= pmax)
          syntax_error (FTL);
        AuxVal = float_node (atof (InputText));
        return (FLOAT);
    }

    switch (c) {
        case EOF:
            return (END);
        case '.':
            return (follow ('.', TO, END));
        case '>':
            return (follow ('=', GE, GT));
        case '<':
            return (follow ('-', ASSN, follow ('=', LE, LT)));
        case '~':
            return (follow ('=', NEQ, NOT));
        case BS:
            return (follow ('/', OR, ERROR));
        case ':':
            return (follow ('=', UASSN, DO));
        case '/':
            return (follow (BS, AND, DIV));
        case '*':
            return (follow ('*', POWER, MULT));
        case '+':
            return (PLUS);
        case '-':
            return (MINUS);
        case '=':
            return (EQ);
        case ';':
            return (SEMI);
        case '(':
            return (LPAREN);
        case ')':
            return (RPAREN);
        case '{':
            return (LBRACE);
        case '}':
            return (RBRACE);
        case '[':
            return (LBRACK);
        case ']':
            return (RBRACK);
        case '|':
            return (BAR);
        case ',':
            return (COMMA);
        case '@':
            return (REFER);
        case '#':
            return (HISTORY);
        case '!':
	    AuxVal = LastCommand;
            return (follow ('!', END, DEREF));
        case '?':
            return (HELP);
        default:
            break;
    }
    return (ERROR);
}

int isnextchar (expect)
int expect;
{
    int     c;
    if ((c = getc (InputFile)) == expect) {
        return (T);
    }
    else {
        ungetc (c, InputFile);
        return (F);
    }
}

int follow (expect, ifyes, ifno)
int expect, ifyes, ifno;
{
    int     c = getc (InputFile);
    if (c == expect) {
        InputText[1] = 'c';
        InputText[2] = '\0';
        return (ifyes);
    }
    ungetc (c, InputFile);
    return (ifno);
}

int backslash(c)
int c;
{
    if (c == BS) {
        switch (c = getc (InputFile)) {
            case 'n':
                return ('\n');
            case 't':
                return ('\t');
            case 'b':
                return ('\b');
            case 'r':
                return ('\r');
            case 'f':
                return ('\f');
            default:
                return (c);
        }
    }
    return (c);
}

NODEPTR symballoc()           /* allocate a new symbol */
{
    if (SymbolTableTop >= SymbolTableMax)
        more_symbol_table (MAXSYMBOLTABLESIZE);
    return (SymbolTableTop++);
}

unsigned hash(s)                         /* form hash value for string s */
char *s;
{
    register unsigned hashval;

    for (hashval = 0; *s != '\0'; )
        hashval = (hashval << 1) ^ (*s++);
    return(hashval % HASHTABLESIZE);
}

NODEPTR lookup(s)
char *s;
{
    register NODEPTR sp;
    for (sp = HashTable[hash (s)]; sp != NULL; sp = node_next(sp))
        if (strcmp (s, node_sym_name(sp)) == 0)
            return (sp);
    return (NULL);
}

NODEPTR install(type,name)
int type;
char *name;
{
    register NODEPTR sp;
    unsigned hashval;
    sp = symballoc ();
    node_type(sp) = type;
    node_sym_name(sp) = name;
    hashval = hash (node_sym_name(sp));
    node_next(sp) = HashTable[hashval];
    HashTable[hashval] = sp;
    return (sp);
}


void install_keyword(type,name)
int type;
char *name;
{
    install (type, name);
}
/*
uninstall(s)
char *s;
{
    SYMBOLPTR * spp;
    for (spp = &HashTable[hash (s)]; *spp != NULL; spp = &((*spp) -> next)) {
        if (strcmp (s, (*spp) -> name) == 0) {
            *spp = (*spp) -> next;
            return (T);
        }
    }
    return(F);
}
*/

/* End of lexer.c. */
