/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987, 1988, 1989, 1990, 1991, 1992, 1993, 1994, 1995,
 * 1996, 1997, 1998, 1999, 2000, 2001, 2002, 2003, 2004, 2005, 2006, 
 * 2007, 2008, 2009, 2010, 2011, 2012  Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: parser.c,v 2.18 2011/10/03 14:14:30 acau Exp $
 */

#include "system.h"


NODEPTR parse(stream)
FILE *stream;
{
    NODEPTR p;
    InputFile = stream;
    SyntaxErrorCount = 0;
    Recover = F;
    next_token ();
    p = resolve_refs (rexpr (0));
    while (Token != END) {
        syntax_error (END);
        next_token ();
    }
    return (SyntaxErrorCount == 0 ? p : True);
}

NODEPTR resolve_refs (e)
NODEPTR e;
{
    if (e == NULL || e == LastCommand)
        return (e);
    switch (node_type(e)) {
        case NAME:
	    return (lookup_env (node_sym_name(e)));
        case NIL:
        case BOOL:
        case INT:
        case STRING:
        case FLOAT:
        case LOC:
            return (e);
        case EXISTS:
            push_scope1 (node_lnk_1(e),node_ival(node_lnk_4(e)));
	    node_lnk_3(e) = resolve_refs (node_lnk_3(e));
	    node_lnk_2(e) = resolve_refs (node_lnk_2(e));
	    node_lnk_1(e) = pop_scope ();
	    return (e);
        case FORALL:
        case FORIN:
        case FOR:
	    push_scope1 (node_lnk_1(e),1);
	    node_lnk_3(e) = resolve_refs (node_lnk_3(e));
	    node_lnk_2(e) = resolve_refs (node_lnk_2(e));
	    node_lnk_1(e) = pop_scope ();
	    return (e);
        case LAMBDA:
	    push_scope1 (node_lnk_1(e),1);
	    node_lnk_2(e) = resolve_refs (node_lnk_2(e));
	    node_lnk_1(e) = pop_scope ();
	    return (e);
        case FORTIMES:
        case IF:
        case SUBLIST:
            node_lnk_3(e) = resolve_refs (node_lnk_3(e));
        case PROJ:
        case COMMA:
        case DEFINE:
        case WHILE:
        case REPEAT:
        case LT:
        case LE:
        case EQ:
        case GE:
        case GT:
        case NEQ:
        case PLUS:
        case MINUS:
        case MULT:
        case DIV:
        case POWER:
        case SEMI:
        case WHERE:
        case AND:
        case OR:
        case IMPLIES:    
        case GETS:
        case IS:
        case MOD:
        case ATANTWO:
        case ASSN:
        case UASSN:
        case CALL:
        case SUB:
        case FIXLIST:
        case VARLIST:
        case NPREV:
        case RUN:
        case SET:
        case SHOWMEM:
        case DEFAULTS:
            node_lnk_2(e) = resolve_refs (node_lnk_2(e));
        case NOT:
        case UPLUS:
        case UMINUS:
        case CEIL:
        case FLOOR:
        case SQRT:
        case ITOF:
        case EXPO:
        case LOG:
        case LOGTEN:
        case SIN:
        case COS:
        case TAN:
        case ASIN:
        case ACOS:
        case ATAN:
        case SINH:
        case COSH:
        case TANH:
        case FABS:
        case PROCESS:
        case PREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case INIT:
        case FIN:
        case CHOPSTAR:    
        case KEEP:    
        case NEXT:
        case WNEXT:
        case PREV:
        case LEN:
        case STB:
        case MEM:
        case STRUCT:
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case PEEK:
        case HELP:
        case SYSTEM:
        case TYPE:
        case iUNDEFINED:
        case ASCII:
        case FORMAT:
        case SIZE:
        case MAKELIST:
        case REFER:
        case DEREF:    
        case LOAD:
            node_lnk_1(e) = resolve_refs (node_lnk_1(e));
        default:
            return (e);
    }
}

NODEPTR rexpr(n)
int n;
{
    NODEPTR p = rbexpr ();
    unsigned tk;
    int lp = 0;
    int rp = 0;
    do {
        switch (tk = Token)  {
            case LPAREN:
                p = rcall (p); 
                continue;
            case LBRACK:
                p = rsublist (p);
                continue;
            case WHERE:
                lp = 36;
                rp = 15;
/*                lp = rp = 15;*/
                break;
            case SEMI:
            case PROJ:
                lp = rp = 20;
                break;
            case IMPLIES:
                lp = rp = 22;
                break;    
            case OR:
                lp = rp = 25;
                break;
            case AND:
                lp = rp = 30;
                break;
            case GETS:
            case IS:
            case ASSN:
            case UASSN:
            case DEFAULTS:
                lp = 50; rp = 49;
                break;
            case EQ:
            case LT:
            case LE:
            case GE:
            case GT:
            case NEQ:
                lp = 55; rp = 54;
                break;
            case PLUS:
            case MINUS:
                lp = rp = 60;
                break;
            case MULT:
            case DIV:
            case MOD:
                lp = rp = 65;
                break;
            case POWER:
                lp = rp = 70;
                break;
            default:
                return (p);
        }
        Recover = F;
        if (n >= lp)
            return (p);
        next_token ();
        p = node2 (tk, p, rexpr (rp));
    } while (T);
}

NODEPTR rbexpr()
{
    NODEPTR p = AuxVal;
    unsigned tk;
    if ((tk = Token) == END)
        return (p);
    next_token ();
    switch (tk) {
        case LPAREN:
            p = rexpr (0);
            assert_syntax (RPAREN);
            return (p);
        case LBRACE:
            p = rexpr (0);
            assert_syntax (RBRACE);
            return (p);
        case NIL:
            return (Nil);
        case TRUE:
            return (True);
        case FALSE:
            return (False);
        case INT:
        case STRING:
        case FLOAT:
        case NAME:
	case PASSLIMIT:
	case VERBOSITY:
	case PRINTLEVEL:
	case DEBUGLEVEL:
	case STATEPRINT:
	case REDUNDANCY:
	case CALLBYNAME:
	case ASSIGNAHEAD:
        case BREAKISABORT:
	case PATH:
	case HELPDIR:
        case INFILE:
        case OUTFILE:
        case HOPCHOP:
	case TIMESTEP:
            return (p);
        case LBRACK:
            return (rlist ());
        case NPREV:
            return (rnprev (tk));
        case FIXLIST:
        case VARLIST:
            return (rlistdef (tk));
        case EMPTY:
            return (Empty);
        case MORE:
            return (More);       /* AC added 05/09/2001 */    
        case SKIP:
            return (Skip);
        case RANLEN:
            return (Ranlen);
        case INFINITE:
            return (Infinite);   
        case RANDOM:             /* Added AC 29Apr1999 */
            return (Random);   
        case RRANDOM:            /* Added AC 20/09/2000 */
            return (RRandom);
        case SRANDOM:
            return (SRandom);
        case SRRANDOM:
            return (SRRandom);    
        case EXISTS:
            return (rexists1 ());
        case FORALL:
            return (rforall ());
        case IF:
            return (rif ());
        case WHILE:
            return (rwhile ());
        case REPEAT:
            return (rrepeat ());
        case CHOPSTAR:
            return (rchopstar ());
        case KEEP:
            return (rkeep ());    
        case FOR:
            return (rfor ());
        case LAMBDA:
            return (rlambda ());
        case DEFINE:
            return (rdefine ());
        case SET:
            return (rset ());
        case FORMAT:
            return (rformat ());
        case ATANTWO:
            return (ratantwo ());
        case RUN:
            return (node2 (tk, rexpr (5), True));
        case PROCESS:
        case PREFIX:
        case ALWAYS:
        case SOMETIME:
        case HALT:
        case FIN:
        case INIT:
        case STB:
        case MEM:
        case LEN:
            return (node1 (tk, rexpr (40)));
        case INPUT:
        case WINPUT:
        case OUTPUT:
        case PEEK:
            return (node1 (tk, rexpr (45)));
        case NOT:
        case NEXT:
        case WNEXT:
        case PREV:
        case TYPE:
        case iUNDEFINED:
        case ASCII:
        case STRUCT:
        case LOAD:
        case SYSTEM:
            return (node1 (tk, rexpr (75)));
        case MINUS:
            return (node1 (UMINUS, rexpr (75)));
        case PLUS:
            return (node1 (UPLUS, rexpr (75)));
        case CEIL:
            return (node1 (CEIL, rexpr (75)));
        case FLOOR:
            return (node1 (FLOOR, rexpr (75)));
        case SQRT:
            return (node1 (SQRT, rexpr (75)));
        case ITOF:
            return (node1 (ITOF, rexpr (75)));
        case EXPO:
            return (node1 (EXPO, rexpr (75)));
        case LOG:
            return (node1 (LOG, rexpr (75)));
        case LOGTEN:
            return (node1 (LOGTEN, rexpr (75)));
        case SIN:
            return (node1 (SIN, rexpr (75)));
        case COS:
            return (node1 (COS, rexpr (75)));
        case TAN:
            return (node1 (TAN, rexpr (75)));
        case ASIN:
            return (node1 (ASIN, rexpr (75)));
        case ACOS:
            return (node1 (ACOS, rexpr (75)));
        case ATAN:
            return (node1 (ATAN, rexpr (75)));
        case SINH:
            return (node1 (SINH, rexpr (75)));
        case COSH:
            return (node1 (COSH, rexpr (75)));
        case TANH:
            return (node1 (TANH, rexpr (75)));
        case FABS:
            return (node1 (FABS, rexpr (75)));    
        case REFER:
        case DEREF:
            return (node1 (tk, rexpr (80)));
        case BAR:
            p = rexpr (0);
            assert_syntax (BAR);
            return (node1 (SIZE, p));
        case SHOWMEM:
            return (node2(SHOWMEM, rexpr(0), rexpr(0)));
                              /* AC 1999-12-01: with this showmem works */
        case EXIT:
        case ABORT:
        case BREAK:
        case RETURN:
        case LEAP:
        case STEP:
        case SHOWSTATE:
        case CONTINUE:
        case SHOWTRACE:
            return (p);
        case HELP:
            return (rhelp ());
        default:
            syntax_error (-1);
            return (True);
    }
}
/*
NODEPTR rexists()
{
    NODEPTR e1,e2;
    NODEPTR nl = rnamelist ();
    switch (Token) {
    case LT:
        if (nl == NULL || node_type(nl) == COMMA)
	    syntax_error(N1);
        next_token ();
        e1 = rexpr (0);
        assert_syntax (DO);
        e2 = rexpr (35);
        return (node3 (EXISTS, nl, e1, e2));
    case DO:
        next_token ();
        e2 = rexpr (35);
        return (node3 (EXISTS, nl, NULL, e2));
    default:
        syntax_error (-1);
        return (True);
    }
}
*/

NODEPTR rexists1()
{
    NODEPTR e1,e2,e3;
    NODEPTR nl = rnamelist ();
    switch (Token) {
    case LT:
        if (nl == NULL || node_type(nl) == COMMA)
	    syntax_error(N1);
        next_token ();
        e1 = rexpr (0);
        if (Token == HISTORY) {
           next_token ();
           e3 = rexpr (0);
        }
        else
            e3 = One;
        if (node_ival(e3)<1 || node_type(e3)!= INT)
            syntax_error(H1);
        if (DebugLevel>2)
            printf("LT: history = %ld\n",node_ival(e3));
        assert_syntax (DO);
        e2 = rexpr (35);
        return (node4 (EXISTS, nl, e1, e2, e3));
    case HISTORY:
        next_token ();
        e3 = rexpr (0);
        if (node_ival(e3)<1 || node_type(e3)!= INT)
            syntax_error(H1);
        if (DebugLevel>2)
            printf("H: history = %ld\n",node_ival(e3));
        assert_syntax (DO);
        e2 = rexpr (35);
        return (node4 (EXISTS, nl, NULL, e2, e3));
    case DO:
        next_token ();
        e2 = rexpr (35);
        e3 = One;
        if (DebugLevel>2)
            printf("DO: history = %ld\n",node_ival(e3));
        return (node4 (EXISTS, nl, NULL, e2, e3));
    default:
        syntax_error (-1);
        return (True);
    }
}


NODEPTR rforall()
{
    NODEPTR e1,e2;
    NODEPTR id = rname ();
    switch (Token) {
    case LT:
        next_token ();
        e1 = rexpr (0);
        assert_syntax (DO);
        e2 = rexpr (35);
        return (node3 (FORALL, id, e1, e2));
    default:
        syntax_error (-1);
        return (True);
    }
}

NODEPTR rif()
{
    NODEPTR b, e1, e2;
    b = rexpr (0);
    assert_syntax (THEN);
    e1 = rexpr (35);
    if (Token == ELSE) {
        next_token ();
        e2 = rexpr (35);
    }
    else
        e2 = True;
    return (node3 (IF, b, e1, e2));
}

NODEPTR rwhile()
{
    NODEPTR b;
    b = rexpr (0);
    assert_syntax (DO);
    return (node2 (WHILE, b, rexpr (35)));
}

NODEPTR rrepeat()
{
    NODEPTR e;
    e = rexpr (35);
    assert_syntax (UNTIL);
    return (node2 (REPEAT, e, rexpr (0)));
}

NODEPTR rchopstar()   
{
    return (node1 (CHOPSTAR, rexpr (35)));
}

NODEPTR rkeep()   
{
    return (node1 (KEEP, rexpr (35)));
}

NODEPTR rfor()
{
    NODEPTR e1, e2;
    e1 = rexpr (0);
    switch (node_type(e1)) {
    case LT:
        assert_syntax (DO);
        return (node3 (FOR, node_lnk_1(e1), node_lnk_2(e1), rexpr (35)));
    default:
        switch (Token) {
        case IN:
            next_token ();
            e2 = rexpr (0);
            assert_syntax (DO);
            return (node3 (FORIN, e1, e2, rexpr (35)));
        case TIMES:
            next_token ();
	    assert_syntax (DO);
            return (node2 (FORTIMES, e1, rexpr (35)));
        default:
            syntax_error (-1);
            return (True);
        }
    }
}

NODEPTR rdefine()
{
    NODEPTR e;
    NODEPTR nl;
    NODEPTR id = rname ();
    if (Token == LPAREN) {
        assert_syntax (LPAREN);
        nl = rnamelist ();
        assert_syntax (RPAREN);
        assert_syntax (EQ);
        e = rexpr (35);
/*        e = rexpr (10);*/
        return (node2 (DEFINE, id, node2 (LAMBDA, nl, e)));
    }
    else {
        assert_syntax (EQ);
        e = rexpr (35);
/*        e = rexpr (10);*/
        return (node2 (DEFINE, id, e));
    }
}

NODEPTR rset()
{
    NODEPTR id = rsysvar ();
    assert_syntax (EQ);
    return (node2 (SET, id, rexpr (35)));
}

NODEPTR rlambda()
{
    NODEPTR e;
    NODEPTR nl;
    assert_syntax (LPAREN);
    nl = rnamelist ();
    assert_syntax (RPAREN);
    switch (Token) {
    case DO:
        next_token ();
        e = rexpr (35);
        return (node2 (LAMBDA, nl, e));
    default:
        syntax_error (-1);
        return (True);
    }
}

NODEPTR rlist()
{
    NODEPTR e;
    if (Token == RBRACK)
        e = NULL;
    else
        e = rexprlist ();
    assert_syntax (RBRACK);
    return (node1 (MAKELIST, e));
}

NODEPTR rsublist(list)
NODEPTR list;
{
    NODEPTR e1, e2;
    assert_syntax (LBRACK);
    e1 = rexpr (0);
    if (Token == TO) {
        next_token ();
        e2 = rexpr (0);
        assert_syntax (RBRACK);
        return (node3 (SUBLIST, list, e1, e2));
    }
    else {
        assert_syntax (RBRACK);
        return (node2 (SUB, list, e1));
    }
}

NODEPTR rlistdef(type)
unsigned int type;
{
    NODEPTR e1, e2;
    assert_syntax (LPAREN);
    e1 = rexpr (0);
    assert_syntax (COMMA);
    e2 = rexpr (0);
    assert_syntax (RPAREN);
    return (node2 (type, e1, e2));
}

NODEPTR rnprev(type)
unsigned int type;
{
    NODEPTR e1, e2;
    assert_syntax (LPAREN);
    e1 = rexpr (0);
    assert_syntax (COMMA);
    e2 = rexpr (0);
    assert_syntax (RPAREN);
    return (node2 (type, e1, e2));
}

NODEPTR rformat()
{
    NODEPTR e;
    assert_syntax (LPAREN);
    e = rexprlist ();
    assert_syntax (RPAREN);
    return (node1 (FORMAT, e));
}

NODEPTR ratantwo()
{
    NODEPTR e1, e2;
    assert_syntax (LPAREN);
    e1 = rexpr (0);
    assert_syntax (COMMA);
    e2 = rexpr (0);
    assert_syntax (RPAREN);
    return (node2 (ATANTWO, e1, e2));
}

NODEPTR rhelp()
{
    NODEPTR s;
    switch (Token) {
        case STRING:
            s = AuxVal;
            break;
        case NAME:
            s = string_node (node_sym_name(AuxVal));
            break;
        case END:
            s = NULL;
            break;
        default:
            s = string_node (strsave (Keyword[Token]));
    }
    while (Token != END)
        next_token ();
    return (node1 (HELP, s));
}

NODEPTR rcall(id)
NODEPTR id;
{
    NODEPTR e;
    assert_syntax (LPAREN);
    if (Token != RPAREN)
        e = rexprlist ();
    else
        e = NULL;
    assert_syntax (RPAREN);
    return (node2 (CALL, id, e)); 
}

NODEPTR rnamelist()
{
    NODEPTR nl ;
    if (Token == NAME) {
	NODEPTR id = rname ();
        if (Token != COMMA)
            return (id);
        next_token ();
        nl = rnamelist ();
        if (nl != NULL)
	  return (node2 (COMMA, id, nl));
        else
            syntax_error(NLC);
    }
    return (NULL);
}

NODEPTR rname()
{
    NODEPTR id = AuxVal;
    assert_syntax (NAME);
    return (id);
}

NODEPTR rexprlist()
{
    NODEPTR e = rexpr (0);
    if (Token != COMMA)
        return (node2 (COMMA, e, NULL));
    next_token ();
    return (node2 (COMMA, e, rexprlist ()));
}

NODEPTR rsysvar()
{
    NODEPTR id = AuxVal;
    switch (Token) {
	case PASSLIMIT:
	case VERBOSITY:
	case PRINTLEVEL:
	case DEBUGLEVEL:
	case STATEPRINT:
	case REDUNDANCY:
	case CALLBYNAME:
	case ASSIGNAHEAD:
        case BREAKISABORT:
	case PATH:
	case HELPDIR:
        case INFILE:
        case OUTFILE:
        case HOPCHOP:
	case TIMESTEP:
        case RANDSEED:
        case RRANDSEED:
        case SRANDSEED:
        case SRRANDSEED:    
	case MAXRANDLEN:
        case PRECISION:
	    next_token ();
	    return (id);
	default:
	    syntax_error (-1);
	    return (NULL);
    }
}

void push_scope(name_list)
NODEPTR name_list;
{
    unsigned int level = node_env_level(EnvStack[EnvTop].env) + 1;
    if (EnvTop >= MAXENVSTACKSIZE)
         error_break0("environment stack full");
    else {
        EnvTop += 1;
        EnvStack[EnvTop].env = env_node(level, 0, NULL);
        EnvStack[EnvTop].env_base = NULL;
        install_name_list (name_list);
    }
}

void push_scope1(name_list,history)
NODEPTR name_list;
unsigned int history;
{
    unsigned int level = node_env_level(EnvStack[EnvTop].env) + 1;
    if (EnvTop >= MAXENVSTACKSIZE)
         error_break0("environment stack full");
    else {
        EnvTop += 1;
        if (history == 0)
            history=1;
        EnvStack[EnvTop].env = env_node1(level, 0, NULL, history);
        EnvStack[EnvTop].env_base = NULL;
        install_name_list (name_list);
    }
}

NODEPTR pop_scope()
{
    return (EnvStack[EnvTop--].env);
}

void install_name_list(name_list)
NODEPTR name_list;
{
    NODEPTR id;
    if (name_list == NULL)
        return;
    while (node_type(name_list) == COMMA) {
	id = node_lnk_1(name_list);
	install_local_id (node_sym_name(id));
	name_list = node_lnk_2(name_list);
    }
    install_local_id (node_sym_name(name_list));
}

NODEPTR lookup_env(s)
char *s;
{
    register NODEPTR np;
    register int level;

    for (level = EnvTop; level >= EnvBase; level--)
        for (np = node_env_idlist(EnvStack[level].env);
	     np != NULL; np = node_next(np))
            if (strcmp (s, node_nm_name(np)) == 0)
                return (np);
    return (install_global_id (s));
}

NODEPTR install_global_id(name)
char *name;
{
    NODEPTR *np; 
    register    NODEPTR id;
    np = &(node_env_idlist(EnvStack[EnvBase].env));
    node_env_history(EnvStack[EnvBase].env)=1;
    if (DebugLevel>2)
        printf("Global: history=%d\n",node_env_history(EnvStack[EnvBase].env));
    id = name_node ((unsigned)isupper ((char)*name), name, 0, offset2 (GlobalMemUsed++,node_env_history(EnvStack[EnvBase].env)));
    if (GlobalMemUsed >= GLOBALMEMSIZE)
        error_break0("global memory full");
    else
	new_loc1 (mem_loc (id), node_type(id), 1);
    while (*np != NULL) {
        np = &(node_next(*np) );
    }
    return (*np = id);
}

void install_local_id(name)
char *name;
{
    NODEPTR *np; 
    register    NODEPTR id;
    np = &(node_env_idlist(EnvStack[EnvTop].env));
    if (DebugLevel>2)
        printf("Local: history=%d\n",node_env_history(EnvStack[EnvTop].env));
    id = name_node ((unsigned)isupper ((char)*name), name, node_env_level(EnvStack[EnvTop].env),
    		    offset2 (node_env_size(EnvStack[EnvTop].env)++, node_env_history(EnvStack[EnvTop].env)));
    while (*np != NULL) {
        np = &(node_next(*np) );
    }
    *np = id;
} 

void next_token()
{
    Token = read_token ();
}

void assert_syntax(tk)
int tk;
{
    if ((int)Token != tk)
        syntax_error (tk);
    else
        next_token ();
}

void syntax_error(n)
int n;
{
    if (Recover)
        return;
    Recover = T;
    SyntaxErrorCount += 1;
    printf ("***Syntax Error");
    if (InputFile != stdin)
        printf (" on line %d of \"%s\"", InputLineCount, InputFilename);
    printf (": \n");
    fflush(stdout);
    syntax_msg (n);
    printf (".\n");
    fflush(stdout);
}

void syntax_msg(n)
int n;
{
    switch (n) {
        case N1:
            printf ("one identifier expected after quantifier");
            break;
        case MQS:
            printf ("missing quote in string");
            break;
        case MDF:
            printf ("missing $ in float");
            break;
        case EIF:
            printf ("error in float");
            break;    
        case STL:
            printf ("string too long");
            break;
        case FTL:
            printf ("float too long");
            break;
        case H1:
            printf ("history should be an integer greater than zero");
            break;     
        case IDN:
            printf ("syntax error, comment doesn't end properly");
            break;
        case NLC:
            printf ("syntax error, comma without a following identifier");
            break;
        case END:
            printf ("\"%s\" unexpected (command must be terminated by \".\")",
                                                                   InputText);
            break;
        default:
            printf ("\"%s\" unexpected", InputText);
            if (0 <= n && n <= MAXTOKEN)
                printf (" (\"%s\" expected)", Keyword[n]);
    }
}

